from setuptools import setup

setup(name='pypersist',
      version='0.2',
      description='Persistent memoisation framework for Python',
      url='https://github.com/mtorpey/pypersist',
      author='Michael Torpey',
      author_email='mct25@st-andrews.ac.uk',
      license='GPL',
      packages=['pypersist'],
      zip_safe=False)
